/*-
 * Copyright (c) 2007-2008 Michael Telahun Makonnen <mtm@FreeBSD.Org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Copyright (c)2004 The DragonFly Project.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 * 
 *   Neither the name of the DragonFly Project nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

/*
 * fn_install.c
 * Installer Function : Install OS Files.
 * $Id: fn_install.c,v 1.74 2006/04/18 19:43:48 joerg Exp $
 */

#include <sys/param.h>
#include <sys/mount.h>
#include <sys/queue.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <sys/wait.h>

#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <libutil.h>
#include <string.h>
#include <unistd.h>

#define SOURCES_CONF_FILE "/usr/local/share/dfuibe_installer/sources.conf"

#ifdef ENABLE_NLS
#include <libintl.h>
#define _(String) gettext (String)
#else
#define _(String) (String)
#endif

#include "aura/mem.h"
#include "aura/buffer.h"
#include "aura/fspred.h"
#define NEEDS_DFUI_STRUCTURE_DEFINITIONS
#include "dfui/dfui.h"
#undef NEEDS_DFUI_STRUCTURE_DEFINITIONS
#include "dfui/system.h"
#include "installer/cmdnames.h"
#include "installer/commands.h"
#include "installer/conf.h"
#include "installer/confed.h"
#include "installer/diskutil.h"
#include "installer/functions.h"
#include "installer/uiutil.h"

#include "flow.h"
#include "pathnames.h"
#include "fn.h"
#include "dist.h"

#define DIST_MNT		DEFAULT_DESTDIR "dist"
#define DIST_CDROMDEV		"/dev/acd0"
#define INFPATH			DIST_MNT "/cdrom.inf"
#define CDVER			"CD_VERSION"

static pid_t	 chroot_chores(struct i_fn_args *);
static void	 enable_swap(struct i_fn_args *);
static int	 do_extract(struct i_fn_args *, const struct be_dist *,
		    const char *);
static int	 do_mount(const char *, const char *, const char *, int);
static struct cf_dptr *find_distptr(struct cf_dptrq *, const char *);
static char	*get_version(struct i_fn_args *);
static int	 mount_cdrom(struct i_fn_args *);
static int	 mount_subpart(struct i_fn_args *);
static int	 skip_dist(struct cf_dptrq *, struct be_dist *, const char *);

static int
do_mount(const char *dev, const char *mnt, const char *type, int flags)
{
	struct iovec iov[6];

	iov[0].iov_base = "fstype";
	iov[0].iov_len = strlen(iov[0].iov_base) + 1;
	iov[1].iov_base = (void *)type;
	iov[1].iov_len = strlen(iov[1].iov_base) + 1;
	iov[2].iov_base = "fspath";
	iov[2].iov_len = strlen(iov[2].iov_base) + 1;
	iov[3].iov_base = (void *)mnt;
	iov[3].iov_len = strlen(iov[3].iov_base) + 1;
	iov[4].iov_base = "from";
	iov[4].iov_len = strlen(iov[4].iov_base) + 1;
	iov[5].iov_base = (void *)dev;
	iov[5].iov_len = strlen(iov[5].iov_base) + 1;
	return (nmount(iov, 6, flags));
}

/*
 * fn_install_os: actually put FreeBSD on a disk.
 */
void
fn_install_os(struct i_fn_args *a)
{
	struct subpartition *sp;
	struct commands *cmds;
	struct command *cmd;
	struct be_dist *dist, *subdist;
	struct cf_dptr *cd;
	char devfsPath[16];
	char dev[MAXPATHLEN];
	char mnt[MAXPATHLEN];
	char fstab[MAXPATHLEN];
	char *cdver;
	FILE *ff;
	long mfsbacked_size;
	pid_t pid;
	int skipped, dist_deselected, cdmounted, fstatus, error;

	cdmounted = 0;
	a->result = 0;
	enable_swap(a);

	if (mount_subpart(a) != 0)
		goto dismount;

	if (mount_cdrom(a) == 0)
		cdmounted = 1;
	else
		goto dismount;

	cdver = get_version(a);
	if (cdver == NULL)
		goto dismount;

	skipped = 0;
	TAILQ_FOREACH(dist, &a->beconf->bc_distq, bd_link) {

		/* Check that this isn't a distribution that should be skipped. */
		if (skip_dist(&a->cc->cc_dptrq, dist, (const char *)a->arch))
			continue;

		/*
		 * Do the actual extraction. A -1 return error means the user
		 * cancelled the whole thing because of an error. Otherwise,
		 * continue to the next distribution, increasing skip count if
		 * necessary.
		 */
		if (TAILQ_EMPTY(&dist->bd_distq)) {
			error = do_extract(a, (const struct be_dist *)dist,
			    (const char *)cdver);
			if (error == 1)
				skipped++;
			else if (error == -1)
				goto dismount;
		} else {
			/*
			 * This is a container distribution.
			 * XXX - This assumes only one level of subdists.
			 */
			TAILQ_FOREACH(subdist, &dist->bd_distq, bd_link) {
				if (skip_dist(&a->cc->cc_dptrq, subdist,
				    (const char *)a->arch))
					continue;
				error = do_extract(a,
				    (const struct be_dist *)subdist,
				    (const char *)cdver);
				if (error == 1)
					skipped++;
				else if (error == -1)
					goto dismount;
			}
		}
	}

	/*
	 * If we were not able to install some distributions let
	 * the user know that it may cause problems, but don't
	 * fail the installation.
	 * Note: the installation may fail further on for other reasons.
	 */
	if (skipped != 0)
		inform(a->c, _("One or more distributions failed to install "
		    "properly. While this may not be a fatal error it may "
		    "cause problems further along in the installation."));

	/*
	 * Write new fstab.
	 */
	snprintf(fstab, sizeof(fstab)/sizeof(char), "%s/etc/fstab",
	    DEFAULT_DESTDIR);
	ff = fopen(fstab, "w");
	if (ff == NULL) {
		inform(a->c, _("Internal Error (%d) - unable to open %s "
		    "for writing"), errno, fstab);
		goto dismount;
	}
	fprintf(ff, "# Device\t\tMountpoint\tFStype\tOptions\t\tDump\tPass#\n");
	for (sp = slice_subpartition_first(storage_get_selected_slice(a->s));
	sp != NULL; sp = subpartition_next(sp)) {
		if (strcmp(subpartition_get_mountpoint(sp), "/") == 0) {
			fprintf(ff, "/dev/%s\t\t%s\t\tufs\trw\t\t1\t1\n",
			    subpartition_get_device_name(sp),
			    subpartition_get_mountpoint(sp));
		} else if (strcmp(subpartition_get_mountpoint(sp), "swap") == 0) {
			fprintf(ff, "/dev/%s\t\tnone\t\tswap\tsw\t\t0\t0\n",
			    subpartition_get_device_name(sp));
		} else if (subpartition_is_mfsbacked(sp)) {
			mfsbacked_size =
			slice_get_capacity(storage_get_selected_slice(a->s)) *
			2048;
			fprintf(ff, "swap\t\t%s\t\t\tmfs\trw,-s%ld\t\t1\t1\n",
			    subpartition_get_mountpoint(sp), mfsbacked_size);
		} else {
			fprintf(ff, "/dev/%s\t\t%s\t\tufs\trw\t\t2\t2\n",
			    subpartition_get_device_name(sp),
			    subpartition_get_mountpoint(sp));
		}
	}
	fclose(ff);

	/*
	 * Mount devfs(5) onto the root patition of the installation..
	 */
	snprintf(devfsPath, sizeof(devfsPath), "%smnt/dev", a->os_root);
	error = do_mount("devfs", devfsPath, "devfs", 0);
	if (error != 0) {
		inform(a->c, _("Internal Error - nmount of devfs on %s failed "
		    "(%d)."), devfsPath, errno);
		goto dismount;
	}

	pid = chroot_chores(a);

	/*
	 * Wait for our chrooted process to finish.
	 */
	fstatus = 0;
	if ((wait4(pid, &fstatus, 0, NULL) < 1) || fstatus != 0) {
		inform(a->c, _("Error - Unable to complete "
		    "post-install fixups."));
		unmount(devfsPath, MNT_FORCE);
		a->result = 0;
		goto dismount;
	} else
		a->result = 1;


	/* Unmount chroot'ed devfs(5) filesystem. */
	unmount(devfsPath, MNT_FORCE);

	cmds = commands_new();

	/*
	 * Clean up.  In case some file didn't make it, use rm -f
	 */
	cmd = command_add(cmds, "%s%s -f %smnt/tmp/install.log",
	    a->os_root, RM, a->os_root);
	command_set_failure_mode(cmd, COMMAND_FAILURE_IGNORE);

	/*
	 * Backup the log. This is not terribly important so we don't
	 * need to inform the user of anything if it fails. It will be
	 * in the log output anyway.
	 */
	cmd = command_add(cmds, "%s%s %sinstall.log %smnt/var/log/install.log",
	    a->os_root, CP,
	    a->tmp, a->os_root);
	command_set_failure_mode(cmd, COMMAND_FAILURE_IGNORE);
	cmd = command_add(cmds, "%s%s 600 %smnt/var/log/install.log",
	    a->os_root, CHMOD, a->os_root);
	command_set_failure_mode(cmd, COMMAND_FAILURE_IGNORE);
	commands_execute(a, cmds);
	commands_free(cmds);

	if (cdmounted) {
		error = unmount(DIST_MNT, MNT_FORCE);
		i_log(a, "%d = unmount(%s)\t\t errno = %d\n",
			error, DIST_MNT, errno);
	}
	return;

	/*
	 * Only reach here in case of some error-condition.
	 */
dismount:
	/*
	 * Unmount everything we mounted.
	 */
	if (cdmounted) {
		error = unmount(DIST_MNT, MNT_FORCE);
		i_log(a, "%d = unmount(%s)\t\t errno = %d\n",
			error, DIST_MNT, errno);
	}
	unmount_subpartitions(a);
}

/*
 * Enable swap partitions
 */
static void
enable_swap(struct i_fn_args *a)
{
	struct subpartition *sp;
	char dev[MAXPATHLEN];
	int  error;

	for (sp = slice_subpartition_first(storage_get_selected_slice(a->s));
	    sp != NULL; sp = subpartition_next(sp)) {
		if (!subpartition_is_swap(sp))
			continue;
		snprintf(dev, sizeof(dev) / sizeof(char), "/dev/%s",
		    subpartition_get_device_name(sp));
		error = swapon(dev);
		i_log(a, "%d = swapon(%s),\t\t errno = %d", error,
		     subpartition_get_device_name(sp), errno);
	}
}

static char *
get_version(struct i_fn_args *a)
{
	properties  infp;
	char	   *cdver;
	const char *cp;
	int	    infd;

	/*
	 * Get the CD version. The path to the directory containing the
	 * distributions is made from this string.
	 */
	infd = open(INFPATH, O_RDONLY);
	if (infd == -1) {
		inform(a->c, _("Internal Error (%d) - Unable to open %s"),
		    errno, INFPATH);
		return (NULL);
	}
	infp = properties_read(infd);
	if (infp == NULL) {
		inform(a->c, _("Internal Error (%d) - Unable to read from %s"),
		    errno, INFPATH);
		return (NULL);
	}
	cp = property_find(infp, CDVER);
	if (cp == NULL) {
		inform(a->c, _("Internal Error - no property: %s"), CDVER);
		return (NULL);
	}
	cdver = strdup((const char *)cp);
	properties_free(infp);
	close(infd);
	i_log(a, "CD Version String: %s", cdver);
	return (cdver);
}

/*
 * Mount release CD-ROM. Returns 0 on success.
 */
static int
mount_cdrom(struct i_fn_args *a)
{
	if (mkdir(DIST_MNT, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) ==
	    -1 && errno != EEXIST) {
		inform(a->c, _("Internal Error (%d) - could not create %s"),
		    errno, DIST_MNT);
		return (-1);
	}
	i_log(a, "Mounting CD-ROM %s on %s", DIST_CDROMDEV, DIST_MNT);
	if (do_mount(DIST_CDROMDEV, DIST_MNT, "cd9660", MNT_RDONLY) != 0) {
		inform(a->c, _("Internal Error (%d)- could not mount CD-ROM. "
		    "Currently, this installer can only mount the first "
		    "IDE CD-CDROM"), errno);
		return (-1);
	}
	return (0);
}

/*
 * Create mount points and mount subpartitions on them.
 * XXX - I assume the root '/' partition is first on the list, so
 *       subsequent mount points are created on top of it.
 *	Returns 0 on success.
 */
static int
mount_subpart(struct i_fn_args *a)
{
	struct subpartition *sp;
	struct commands *cmds;
	struct command *cmd;
	char dev[MAXPATHLEN];
	char mnt[MAXPATHLEN];
	int  error;

	for (sp = slice_subpartition_first(storage_get_selected_slice(a->s));
	     sp != NULL; sp = subpartition_next(sp)) {
		if (subpartition_is_swap(sp) || subpartition_is_mfsbacked(sp))
			continue;
	
		snprintf(dev, sizeof(dev) / sizeof(char), "%sdev/%s",
		    a->os_root, subpartition_get_device_name(sp));
		snprintf(mnt, sizeof(mnt) / sizeof(char), "%smnt%s",
		    a->os_root, subpartition_get_mountpoint(sp));
		error = mkdir(mnt,
		    S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
		if (error == -1 && errno != EEXIST) {
			inform(a->c, _("Internal Error (%d) - "
			    "could not create directory %s"), errno, mnt);
			return (-1);
		}
		i_log(a, "%d = mkdir(%s)\t\t errno = %d", error, mnt, errno);
		i_log(a, "Mounting %s on %s...", dev, mnt);
		if (do_mount(dev, mnt, "ufs", 0) == -1) {
			inform(a->c, _("Internal Error (%d) - unable to "
			    "mount %s on %s"), errno, dev, mnt);
			return (-1);
		}
	}
	return (0);
}

static int
skip_dist(struct cf_dptrq *cdq, struct be_dist *dist, const char *arch)
{
	struct cf_dptr *cd;
	int dist_deselected;

	/*
	 * Skip unselected distributions. Assume it's NOT selected
	 * until we find it in our list of selected distributions.
	 */
	dist_deselected = 1;
	TAILQ_FOREACH(cd, cdq, cd_link)
		if (strcmp(dist->bd_id, cd->cd_id) == 0)
			dist_deselected = 0;
	if (dist_deselected)
		return (1);

	/*
	 * Skip if hardware architecture of distribution does not
	 * match the machine we're running on.
	 */
	if (dist->bd_arch != NULL &&
		strcmp(dist->bd_arch, arch) != 0)
		return (1);

	return (0);
}

/*
 * Extract the given distribution. Returns 0 if the distribution was
 * extracted successfully, 1 if it was skipped, and -1 if it was NOT
 * extracted successfully.
 */
static int
do_extract(struct i_fn_args *a, const struct be_dist *dist, const char *cdver)
{
	struct dfui_progress *pr;
	char distpath[256];
	char prompt[256];
	int cancelled, error;

retry:
	cancelled = 0;
	snprintf(prompt, sizeof(prompt), _("Extracting %s into %s directory"),
	    dist->bd_id, dist->bd_dst);
	pr = dfui_progress_new(dfui_info_new(
	    _("Extracting Distributions"),
	    prompt,
	    ""),
	    0);

	if (!dfui_be_progress_begin(a->c, pr))
		abort_backend();

	snprintf(distpath, 256, "%s/%s/%s", DIST_MNT, cdver, dist->bd_src);
	i_log(a, "Extracting %s from %s to %s", dist->bd_id,
	    distpath, dist->bd_dst);
	error = distExtractTarball(a, distpath, dist->bd_file, dist->bd_dst,
	    pr, &cancelled);
	if (error == 0) {
		if (!dfui_be_progress_end(a->c))
			abort_backend();
		switch (dfui_be_present_dialog(a->c,
			_("Extraction failed!"),
			_("Retry|Skip|Cancel"),
			_("Extraction of %s failed.\n\n"),
			dist->bd_id)) {
		case 1:
			dfui_progress_free(pr);
			goto retry;
			break;
		case 2:
			return (1);
			break;
		case 3:
		default:
			dfui_progress_free(pr);
			return (-1);
			break;
		}
	}

	if (!dfui_be_progress_end(a->c))
		abort_backend();
	dfui_progress_free(pr);
	return (0);
}

/*
 * This function will fork(2) a new process, chroot(2) to the new installation
 * and perform additional installation chores with the root mount of the
 * new installation as its root directory. Returns pid of forked process.
 */
static pid_t
chroot_chores(struct i_fn_args *a)
{
	struct commands *cmds;
	pid_t pid;
	int   err;

	pid = fork();
	if (pid != 0)
		return (pid);

	/*
	 * Child Process
	 */
	err = 0;
	if (chdir(DEFAULT_DESTDIR) != 0)  {
		i_log(a, "Failed to chdir to %s (%d)",
			DEFAULT_DESTDIR, errno);
		goto out;
	}
	if (chroot(DEFAULT_DESTDIR) != 0)
		goto out;

	i_log(a, "chroot'ed to newly installed system\n");
	cmds = commands_new();

	/*
	 * These were obtained from the original sysinstall.
	 * We should probably check to make sure they are still needed.
	 */
	err = mkdir("/var/db", S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	i_log(a, "%d = mkdir(%s)\t\t errno = %d", err, "var/db", errno);

	close (open("/var/db/mounttab", O_CREAT | O_TRUNC | O_WRONLY,
	    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));

	err = mkdir("/usr/compat",
	    S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	i_log(a, "%d = mkdir(%s)\t\t errno = %d", err, "/usr/compat", errno);

	symlink("/usr/compat", "/compat");

	/* Mail aliases. */
	command_add(cmds, "/%s", NEWALIASES);

	/*
	 * Use mtree(8) to fix permissions on installed files/dirs.
	 */
	command_add(cmds, "/%s -deU -f /etc/mtree/BSD.root.dist -p /", MTREE);
	command_add(cmds, "/%s -deU -f /etc/mtree/BSD.var.dist -p /var", MTREE);
	command_add(cmds, "/%s -deU -f /etc/mtree/BSD.usr.dist -p /usr", MTREE);
	command_add(cmds, "/%s -rf /boot/kernel", RM);
	command_add(cmds, "/%s /boot/GENERIC /boot/kernel", MV);
	
	err = commands_execute(a, cmds);
	commands_free(cmds);

out:
	exit(!err);
}

void
unmount_subpartitions(struct i_fn_args *a)
{
	struct subpartition *sp;
	char path[MAXPATHLEN];
	int  error;

	for (sp = slice_subpartition_last(storage_get_selected_slice(a->s));
	     sp != NULL; sp = subpartition_prev(sp)) {
		if (subpartition_is_swap(sp)) {
			snprintf(path, sizeof(path) / sizeof(char), "/dev/%s",
			    subpartition_get_device_name(sp));
			error = swapoff(path);
			i_log(a, "%d = swapoff(%s)\t\t errno = %d\n",
			    error, path, errno);
			continue;
		}
		snprintf(path, sizeof(path) / sizeof(char), "%smnt%s",
		    a->os_root, subpartition_get_mountpoint(sp));
		if (unmount(path, MNT_FORCE) == -1)
			inform(a->c, _("Warning: subpartition %s was not "
			    "correctly unmounted. Error: %d"), path, errno);
	}
}

struct dfui_form *
create_distributions_form(struct i_fn_args *a)
{
	struct dfui_celldata *cell;
	struct dfui_dataset *ds;
	struct dfui_form    *f;
	struct be_dist	    *bd;
	char buf[256];

	snprintf(buf, sizeof(buf) / sizeof(char),
	    _("Each distribution installs a subset of the %s Operating System. "
	    "Select the individual distributions you would like installed.\n\n"),
	    OPERATING_SYSTEM_NAME);

	f = dfui_form_create(
	    "distributions",
	    _("Select Distributions"),
	    buf,

	    "p", "minimum_width", "64",

	    "f", "x", _(" X"), "", "",
	    "p", "control", "checkbox",

	    "f", "id", _(" ID      "), "", "",
	    "p", "editable", "false",

	    "f", "desc", _(" Description                            "), "", "",
	    "p", "editable", "false",

	    "a", "ok", _("Begin Installation"), "", "",
	    "a", "cancel", _("Return to Create Subpartitions"), "", "",
	    "p", "accelerator", "ESC",
	    NULL);

	dfui_form_set_multiple(f, 1);

	TAILQ_FOREACH(bd, &a->beconf->bc_distq, bd_link) {

		/* Respect "arch" attribute */
		if (bd->bd_arch != NULL && strcmp(bd->bd_arch, a->arch) != 0)
			continue;

		ds = dfui_dataset_new();
		if (bd->bd_required ||
		    find_distptr(&a->cc->cc_dptrq, (const char *)bd->bd_id) != NULL)
			dfui_dataset_celldata_add(ds, "x", "Y");
		else
			dfui_dataset_celldata_add(ds, "x", "N");
		dfui_dataset_celldata_add(ds, "id", bd->bd_id);
		dfui_dataset_celldata_add(ds, "desc", bd->bd_desc);
		dfui_form_dataset_add(f, ds);
	}

	return (f);
}

void
save_distributions_selection(struct dfui_response *r, struct i_fn_args *a)
{
	struct dfui_dataset *ds;
	struct cf_dptr	    *cd;
	const char *checked;
	const char *id;

	for (ds = dfui_response_dataset_get_first(r); ds != NULL;
	    ds = dfui_dataset_get_next(ds)) {
		checked = dfui_dataset_get_value(ds, "x");
		id = dfui_dataset_get_value(ds, "id");
		cd = find_distptr(&a->cc->cc_dptrq, id);

		/*
		 * If the user selected a dist. but it isn't in our list
		 * of dptr's, then add it. Otherwise, if it is in our list
		 * but the user has deselected it, remove it from our list. 
		 */
		if ((strcmp(checked, "Y") == 0) && cd == NULL) {
			new_dptr(&cd, 1);
			if (cd != NULL) {
				cd->cd_id = strdup(id);
				if (cd->cd_id != NULL)
					TAILQ_INSERT_TAIL(&a->cc->cc_dptrq, cd,
					    cd_link);
			}
		} else if ((strcmp(checked, "N") == 0) && cd != NULL) {
			TAILQ_REMOVE(&a->cc->cc_dptrq, cd, cd_link);
			destroy_dptr(cd);
		}
	}
}

/*
 * Finds a dist ptr in the installation configuration structure by it's id
 * or by its display name.
 * If it doesn't exist it returns NULL instead.
 */
static struct cf_dptr *
find_distptr(struct cf_dptrq *cdq, const char *id)
{
	struct cf_dptr *cd;

	TAILQ_FOREACH(cd, cdq, cd_link) {
		if (strcmp(cd->cd_id, id) == 0)
			return (cd);
	}

	return (NULL);
}

void
fn_set_hostname(struct i_fn_args *a)
{
	struct commands *cmds;

	cmds = commands_new();
	command_add(cmds, "%s%s hostname='%s' >> %s%s", DEFAULT_DESTDIR, ECHO,
	    a->cc->cc_hostname, DEFAULT_DESTDIR, "etc/rc.conf");
	commands_execute(a, cmds);
}

void
fn_set_rootpassword(struct i_fn_args *a)
{
	pid_t pid;
	int   fds[2];
	int   error, len, status;

	/*
	 * Use the pw(8) command to set the root password. Fork a new process,
	 * change the root directory to our newly installed system and execute
	 * the pw(8) program. Send the password on its standard input, which
	 * is a pipe we create in this process.
	 */
	error = pipe(fds);
	if (error == -1) {
		inform(a->c, _("Internal Error - pipe(2) returned %d."), errno);
		a->result = 0;
		return;
	}

	pid = fork();
	if (pid == 0) {
		/*
		 * Child
		 */
		dup2(fds[0], 0);
		close(fds[1]);
		chroot(DEFAULT_DESTDIR);
		execl("/" PW, "/" PW, "usermod", "-h", "0", NULL);
		i_log(a, "ERROR! execl() of /%s failed!\n", PW);
		exit(1);
	} else {
		close(fds[0]);

		/*
		 * Write the password on pw(8)'s stdin and check its return
		 * status for whether it succeeded or not.
		 */
		status = 0;
		write(fds[1], a->cc->cc_rootpw, strlen(a->cc->cc_rootpw));
		if ((wait4(pid, &status, 0, NULL) < 1) || status != 0) {
			inform(a->c, _("Error - Unable to set root password."));
			a->result = 0;
		} else
			a->result = 1;
	}
}
