/*
 * Copyright (c)2004 The DragonFly Project.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 * 
 *   Neither the name of the DragonFly Project nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

/*
 * survey.c
 * Survey the storage capacity of the system.
 * $Id: survey.c,v 1.18 2005/08/26 22:44:37 cpressey Exp $
 */

#include <sys/types.h>
#if defined(__OpenBSD__) || defined (__NetBSD__)
#include <sys/param.h>
#endif
#include <sys/sysctl.h>

#include <stdio.h>
#include <string.h>

#ifdef SYSTEM_AURA
#include <aura/dict.h>
#else
#include "dict.h"
#endif

#define NEEDS_DISKUTIL_STRUCTURE_DEFINITIONS
#include "commands.h"
#include "diskutil.h"
#undef NEEDS_DISKUTIL_STRUCTURE_DEFINITIONS
#include "functions.h"
#include "sysinstall2.h"

#if defined(__OpenBSD__) || defined(__NetBSD__)
int mib[2];
int i;
#endif

/* Set to 1 to indicate disks have been located and libdisks(3) initialized. */
int g_disks_inited;

static int	init_disks(struct storage *);
static int	refresh_slices(struct si_disk *);

/*
 * Survey storage capacity of this system.
 * Assumes that the number of disks does not change during the
 * lifetime of the program.
 */
int
survey_storage(struct i_fn_args *a)
{
	long mem = 0;
	size_t len;
	int failure = 0;

	if (!g_disks_inited)
		disks_free(a->s);

	/*
	 * Record size of memory (RAM).
	 */
	len = sizeof(mem);
	if (sysctlbyname("hw.physmem", &mem, &len, NULL, 0) < 0)
		failure = 1;
	else
		storage_set_memsize(a->s, next_power_of_two(mem >> 20));

	/*
	 * If this is the first time we are running, initialize both the
	 * disk and slice structure. Otherwise, the disks are already
	 * initialized so refresh only the slices on each disk.
	 */
	if (g_disks_inited)
		failure += refresh_slices(a->s->disk_head);
	else
		failure += init_disks(a->s);

	if (!failure && !g_disks_inited)
		g_disks_inited = 1;

	return(!failure);
}

static int
init_disks(struct storage *s)
{
	struct si_disk *d;
	struct chunk   *sl;
	struct disk    *ld;
	size_t	len;
	char   *disks, *disk, *disk_ptr;
	char	desc[32];
	int	failure, n;

	/*
	 * Get a list of hard disks in the system.
	 */
	failure = 0;
	if (sysctlbyname("kern.disks", NULL, &len, NULL, 0) != -1) {
		/* For terminating NULL */
		len++;
		disks = (char *)malloc(len);
		if (disks != NULL) {
			memset((void *)disks, 0, len);
			if (sysctlbyname("kern.disks", disks, &len, NULL, 0) !=
			    -1)
				disk_ptr = disks;
			else
				failure = 1;
		} else
			failure = 1;
	} else
		failure = 1;

	/*
	 * Iterate over all the disks and record: geometry and slice info.
	 */
	while (!failure && (disk = strsep(&disk_ptr, " ")) != NULL) {
		if (disk[0] == '\0')
			continue;

		/*
		 * If the disk is a memory disk or CD-ROM, skip it.
		 */
		if (strncmp(disk, "md", 2) == 0 ||
		    strncmp(disk, "cd", 2) == 0)
			continue;

		ld = Open_Disk((const char *)disk);
		if (ld != NULL) {
			n = 1;
			d = disk_new(s, ld, disk);
			disk_set_geometry(d, ld->bios_cyl, ld->bios_hd,
			    ld->bios_sect, ld->sector_size);
			snprintf(desc, 32, "%s [%u MB]",
			    disk_get_device_name(d), ld->bios_cyl *
			    ld->bios_hd * ld->bios_sect * ld->sector_size
			    / 1024 / 1024);
			disk_set_desc(d, desc);
			for (sl = ld->chunks->part; sl != NULL; sl = sl->next) {
				slice_new(d, sl, n, sl->subtype, sl->flags,
				    sl->offset, sl->size);
				n++;
			}
		} else
			failure = 1;
	}

	return (failure);
}

static int
refresh_slices(struct si_disk *head)
{
	struct si_disk *d;
	struct chunk   *c;
	int n;

	/*
	 * First, delete all the slices.
	 */
	for (d = head; d != NULL; d = d->next) {
		slices_free(d->slice_head);
		d->slice_head = NULL;
		d->slice_tail = NULL;
	}

	/*
	 * Rescan slices.
	 */
	for (d = head; d != NULL; d = d->next) {
		n = 1;
		for (c = d->libdisk->chunks->part; c != NULL; c = c->next) {
			slice_new(d, c, n, c->subtype, c->flags, c->offset,
			    c->size);
			n++;
		}
	}
	return (0);
}
