/*-
 * Copyright (c) 2007-2008 Michael Telahun Makonnen <mtm@FreeBSD.Org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/errno.h>
#include <sys/queue.h>
#include <bsdxml.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "conf.h"
#include "private.h"

struct xml_tag {
	TAILQ_ENTRY(xml_tag) xt_link;
	XML_Char **xt_attr;
	XML_Char  *xt_name;
};

struct user_data {
	TAILQ_HEAD(xml_head, xml_tag) ud_xmlStack;
	enum xhandler_t	     ud_xh;
	union conf_selector *ud_cs;
	int ud_err;
};

static int	attr_exist(const XML_Char **, const char *);
static char	*attr_value(const XML_Char *, const XML_Char **);
static void	start_tag_common(void *, const XML_Char *, const XML_Char **);
static void	end_tag_common(void *, const XML_Char *);
static void	char_data_common(void *, const XML_Char *, int);
static void	start_tag_desc(void *, const XML_Char *, const XML_Char **);
static void	end_tag_desc(void *, const XML_Char *);
static void	char_data_desc(void *, const XML_Char *, int);
static void	start_tag_config(void *, const XML_Char *, const XML_Char **);
static void	end_tag_config(void *, const XML_Char *);
static void	char_data_config(void *, const XML_Char *, int);
static void	set_dist_attributes(struct be_dist *, const XML_Char **);
static void	inherit_parent_attrs(struct be_dist *, struct be_dist *);
static int	is_whitespace(const XML_Char);
static XML_Char *strip_delimiter_whitespace(const XML_Char *, int *);

/*
 * This function takes a stream to an xml file and a backend configuration
 * structure. It will initialize this structure and set it according to
 * the information in the xml file.
 * Returns 0 on success, or -1 on failure.
 */
int
xml2mem(FILE *stream, union conf_selector *cs, enum xhandler_t xt)
{
	struct user_data ud;
	struct be_conf *beconf;
	struct cf_conf *cfconf;
	enum XML_Status error;
	XML_Parser p;
	char buf[BUFSIZ];
	int done, len;

	beconf = NULL;
	cfconf = NULL;
	memset((void *)&ud, 0, sizeof(struct user_data));
	TAILQ_INIT(&ud.ud_xmlStack);
	if (xt == XT_BE) {
		beconf = (struct be_conf *)cs;
		TAILQ_INIT(&beconf->bc_distq);
		ud.ud_cs = (union conf_selector *)beconf;
		ud.ud_xh = XT_BE;
	} else {
		cfconf = (struct cf_conf *)cs;
		TAILQ_INIT(&cfconf->cc_dptrq);
		ud.ud_cs = (union conf_selector *)cfconf;
		ud.ud_xh = XT_CONFIG;
	}

	/*
	 * Initialize libxml parsers. Set the callback functions
	 * for start and end tags and for character data not in
	 * a tag. Set the data structure shared between handlers.
	 */
	p = XML_ParserCreate(NULL);
	if (p == NULL) {
		warn("XML_ParserCreate() failed");
		return (-1);
	}
	XML_SetElementHandler(p, start_tag_common, end_tag_common);
	XML_SetCharacterDataHandler(p, char_data_common);
	XML_SetUserData(p, (void *)&ud);
	
	done = 0;
	while (!done) {
		len = fread(buf, 1, BUFSIZ, stream);
		if (ferror(stdin)) {
			warn("fread of stream failed");
			return (-1);
		}
		done = feof(stream);
		error = XML_Parse(p, buf, len, done);
		if (error == XML_STATUS_ERROR) {
			warn("XML Parse error: line %d, column %d, %s",
			    XML_GetCurrentLineNumber(p),
			    XML_GetCurrentColumnNumber(p),
			    XML_ErrorString(XML_GetErrorCode(p)));
			return (-1);
		}
		if (ud.ud_err != 0) {
			warnc(ud.ud_err, "handler error");
			return (-1);
		}
	}
	XML_ParserFree(p);
	return (0);
}

static void
start_tag_common(void *p, const XML_Char *name, const XML_Char **attr)
{
	struct user_data *ud;
	struct xml_tag	 *xt;
	int i;

	ud = (struct user_data *)p;
	xt = (struct xml_tag *)malloc(sizeof(struct xml_tag));
	if (xt == NULL) {
		ud->ud_err = ENOMEM;
		return;
	}
	xt->xt_name = (XML_Char *)strdup((const char *)name);
	if (xt->xt_name == NULL) {
		ud->ud_err = ENOMEM;
		return;
	}

	/* count number of attributes */
	for (i = 0; attr[i] != NULL; i++)
		;;

	/* We depend on calloc(3) pre-zeroing attributes list. */
	xt->xt_attr = (XML_Char **)calloc(sizeof(XML_Char **), i == 0 ? 1 : i);
	if (xt->xt_attr == NULL) {
		ud->ud_err = ENOMEM;
		goto errquit;
	}
	for (i = 0; attr[i] != NULL; i++) {
		xt->xt_attr[i] = strdup(attr[i]);
		if (xt->xt_attr[i] == NULL) {
			ud->ud_err = ENOMEM;
			goto errquit;
		}
	}
	xt->xt_attr[i] = NULL;

	/*
	 * Call the appropriate handler.
	 */
	if (ud->ud_xh == XT_BE)
		start_tag_desc(p, name, attr);
	else
		start_tag_config(p, name, attr);

	TAILQ_INSERT_TAIL(&ud->ud_xmlStack, xt, xt_link);
	return;

errquit:
	if (xt->xt_name != NULL)
		free(xt->xt_name);
	if (xt->xt_attr != NULL) {
		for (i = 0; xt->xt_attr[i] != NULL; i++)
			free(xt->xt_attr[i]);
		free(xt->xt_attr);
	}
}

static void
end_tag_common(void *p, const XML_Char *name)
{
	struct user_data *ud;
	struct xml_tag *xt;

	ud = (struct user_data *)p;
	if (ud->ud_err)
		return;

	FIN_ASSERT(!TAILQ_EMPTY(&ud->ud_xmlStack),
	    "xml stack cannot be empty");
	xt = TAILQ_LAST(&ud->ud_xmlStack, xml_head);
	TAILQ_REMOVE(&ud->ud_xmlStack, xt, xt_link);
	free(xt->xt_name);

	/*
	 * Call the appropriate handler after we remove tag from
	 * the xml stack because the start tag handler is called
	 * before the tag is put on the stack.
	 */
	if (ud->ud_xh == XT_BE)
		end_tag_desc(p, name);
	else
		end_tag_config(p, name);
}

static void
char_data_common(void *p, const XML_Char *s, int len)
{
	struct user_data *ud;

	ud = (struct user_data *)p;
	if (ud->ud_err)
		return;

	/*
	 * Call the appropriate handler.
	 */
	if (ud->ud_xh == XT_BE)
		char_data_desc(p, s, len);
	else
		char_data_config(p, s, len);
}

static void
start_tag_desc(void *p, const XML_Char *name, const XML_Char **attr)
{
	struct user_data *ud;
	struct be_conf	 *bc;
	struct be_dist *dd, *ddtmp;
	struct xml_tag	  *xt;

	ud = (struct user_data *)p;
	bc = (struct be_conf *)ud->ud_cs;

	/*
	 * tag: dist, parent: distributions
	 *	or
	 * tag: dist, parent: dist
	 */
	if (STREQUAL(name, TAG_DIST)) {
		new_bedist(&dd, 1);
		if (dd == NULL) {
			ud->ud_err = ENOMEM;
			return;
		}
		set_dist_attributes(dd, attr);

		/*
		 * If the last element we processed still on the stack was
		 * a container distribution, then this is a subdist of it.
		 * Otherwise, it is a top-level distribution.
		 */
		xt = TAILQ_LAST(&ud->ud_xmlStack, xml_head);
		if (STREQUAL(
		    TAILQ_LAST(&ud->ud_xmlStack, xml_head)->xt_name, TAG_DIST) 
		    && attr_exist((const XML_Char **)xt->xt_attr,
		    ATTR_CONTAINER)) {
			ddtmp = TAILQ_LAST(&bc->bc_distq, be_distq);
			inherit_parent_attrs(dd, ddtmp);
			TAILQ_INSERT_TAIL(&ddtmp->bd_distq, dd, bd_link);
		} else
			TAILQ_INSERT_TAIL(&bc->bc_distq, dd, bd_link);
	}
}

static void
end_tag_desc(void *p, const XML_Char __unused *name)
{
	;;
}

static void
char_data_desc(void *p, const XML_Char *s, int len)
{
	struct user_data *ud;
	struct be_conf	 *bc;
	struct xml_tag	 *xt, *xtprev;
	struct be_dist	 *d, *dparent;
	XML_Char *sclean;

	ud = (struct user_data *)p;
	bc = (struct be_conf *)ud->ud_cs;
	xt = TAILQ_LAST(&ud->ud_xmlStack, xml_head);
	if (xt == NULL)
		return;
	xtprev = TAILQ_PREV(xt, xml_head, xt_link);
	if (xtprev == NULL)
		return;

	/*
	 * tag: dist
	 *	or
	 * tag: desc, parent: dist
	 */
	if (STREQUAL(xt->xt_name, TAG_DIST) || (STREQUAL(xt->xt_name, TAG_DESC)
	    && STREQUAL(xtprev->xt_name, TAG_DIST))) {

		/*
		 * There may be several levels of subdistributions, so drill
		 * down to the bottom.
		 */
		d = TAILQ_LAST(&bc->bc_distq, be_distq);
		while (!TAILQ_EMPTY(&d->bd_distq)){
			dparent = d;
			d = TAILQ_LAST(&dparent->bd_distq, be_distq);
		}

		sclean = strip_delimiter_whitespace(s, &len);

		/*
		 * We may be given the contents of an element over more
		 * than one call to this function.
		 */
		if (d->bd_desc != NULL) {
			d->bd_desc = (char *)reallocf(d->bd_desc, strlen(d->bd_desc) *
			    sizeof(XML_Char) + (sizeof(XML_Char) * len) + 1);
			if (d->bd_desc == NULL) {
				ud->ud_err = ENOMEM;
				return;
			}
			strncat(d->bd_desc, s, len);
		} else {
			d->bd_desc = (char *)malloc(sizeof(XML_Char) * (len + 1));
			if (d->bd_desc == NULL) {
				ud->ud_err = ENOMEM;
				return;
			}
			memcpy((void *)d->bd_desc, (void *)s, (size_t)len);
			d->bd_desc[len] = '\0';
		}
	}
}

static void
start_tag_config(void *p, const XML_Char *name, const XML_Char **attr)
{
	struct user_data *ud;
	struct cf_conf	 *cf;
	struct xml_tag	 *xt;
	struct cf_dptr	 *dp;

	ud = (struct user_data *)p;
	cf = (struct cf_conf *)ud->ud_cs;
}

static void
end_tag_config(void *p, const XML_Char __unused *name)
{
	;;
}

static void
char_data_config(void *p, const XML_Char *s, int len)
{
	struct user_data *ud;
	struct cf_conf	 *cc;
	struct cf_dptr	 *cd;
	struct xml_tag	 *xt;
	XML_Char	 *sclean;

	ud = (struct user_data *)p;
	cc = (struct cf_conf *)ud->ud_cs;

	/*
	 * last tag on xml stack: distid
	 */
	xt = TAILQ_LAST(&ud->ud_xmlStack, xml_head);
	sclean = strip_delimiter_whitespace(s, &len);
	if (len < 1)
		return;
	if (STREQUAL(xt->xt_name, TAG_DISTID)) {
		new_dptr(&cd, 1);
		if (cd == NULL) {
			ud->ud_err = ENOMEM;
			return;
		}
		cd->cd_id = (char *)malloc(sizeof(XML_Char) * (len + 1));
		if (cd->cd_id == NULL) {
			free(cd);
			ud->ud_err = ENOMEM;
			return;
		}
		memcpy((void *)cd->cd_id, (void *)sclean, len);
		cd->cd_id[len] = '\0';
		TAILQ_INSERT_TAIL(&cc->cc_dptrq, cd, cd_link);
	} else if (STREQUAL(xt->xt_name, TAG_HOSTNAME)) {
		cc->cc_hostname = (char *)malloc(sizeof(XML_Char) * (len + 1));
		if (cc->cc_hostname == NULL) {
			ud->ud_err = ENOMEM;
			return;
		}
		memcpy((void *)cc->cc_hostname, (void *)sclean, len);
		cc->cc_hostname[len] = '\0';
	} else if (STREQUAL(xt->xt_name, TAG_ROOTPASSWORD)) {
		cc->cc_rootpw = (char *)malloc(sizeof(XML_Char) * (len + 1));
		if (cc->cc_rootpw == NULL) {
			ud->ud_err = ENOMEM;
			return;
		}
		memcpy((void *)cc->cc_rootpw, (void *)sclean, len);
		cc->cc_rootpw[len] = '\0';
	}
}

static XML_Char *
strip_delimiter_whitespace(const XML_Char *s, int *len)
{
	XML_Char *ssub;

	for (ssub = (XML_Char *)s;
	    *len > 0 &&
	    (const XML_Char)is_whitespace(ssub[*len - 1]); (*len)--, ssub++)
		;;
	return (ssub);
}

static int
is_whitespace(const XML_Char c)
{
	char wslist[] = {' ', '\t', '\n', '\r'};
	int j;

	for (j = 0; j < (sizeof(wslist) / sizeof(char)); j++) {
		if (c == wslist[j])
			return (1);
	}
	return (0);
}

static void
set_dist_attributes(struct be_dist *dd, const XML_Char **attr)
{
	int i;

	for (i = 0; attr[i] != NULL; i += 2) {
		if (STREQUAL(attr[i], ATTR_ID)) {
			dd->bd_id = strdup(attr[i + 1]);

			/*
			 * By default the file name is the same as the id.
			 */
			if (dd->bd_file == NULL)
				dd->bd_file = strdup(dd->bd_id);
		}
		if (STREQUAL(attr[i], ATTR_SRC))
			dd->bd_src = strdup(attr[i + 1]);
		if (STREQUAL(attr[i], ATTR_DST))
			dd->bd_dst = strdup(attr[i + 1]);
		if (STREQUAL(attr[i], ATTR_ARCH))
			dd->bd_arch = strdup(attr[i + 1]);
		if (STREQUAL(attr[i], ATTR_REQUIRED))
			dd->bd_required = 1;
		if (STREQUAL(attr[i], ATTR_FILE)) {

			/*
			 * Remove the default display name (if it exists), and
			 * replace it with another string.
			 */
			if (dd->bd_file != NULL)
				free(dd->bd_file);
			dd->bd_file = strdup(attr[i + 1]);
		}
	}
}

/*
 * Some attributes can be inherited from the container distribution. If any
 * of these are unset, then set them from the parent distribution.
 */
static void
inherit_parent_attrs(struct be_dist *d, struct be_dist *parent)
{
	if (d->bd_src == NULL)
		d->bd_src = parent->bd_src;
	if (d->bd_dst == NULL)
		d->bd_dst = parent->bd_dst;
	if (d->bd_arch == NULL)
		d->bd_arch = parent->bd_arch;
}

static int
attr_exist(const XML_Char **attr, const char *srch)
{
	int i;

	for (i = 0; attr[i] != NULL; i += 2) {
		if (STREQUAL(attr[i], srch))
			return (1);
	}
	return (0);
}

static char *
attr_value(const XML_Char *key, const XML_Char **attr)
{
	int i;

	for (i = 0; attr[i] != NULL; i += 2) {
		if (STREQUAL(attr[i], key))
			return ((char *)attr[i + 1]);
	}
	return (NULL);
}
